import threading
import queue

from enum import Enum, auto

import serial
import crcmod

from infupy.backends.common import Syringe, CommandError, printerr

DEBUG = False

crcccitt = crcmod.predefined.mkCrcFun('crc-ccitt-false')
def genCheckSum(msg):
    crcval = crcccitt(msg)
    checkbytes = b'%04X' % crcval
    return checkbytes

def genFrame(msg):
    return b'!' + msg + b'|' + genCheckSum(msg) + b'\r'

def parseReply(rxbytes):
    # The checksum is seperated by a pipe
    rxmsg, chk = rxbytes.split(b'|', 1)

    # Fields are seperated by caret (HL7 style)
    fields = rxmsg.split(b'^')

    ret = fields[1:]
    return (ret, chk == genCheckSum(rxmsg))

class Looper(threading.Thread):
    def __init__(self, syringe, delay, stopevent):
        super().__init__(daemon=True)
        self.stopped = stopevent
        self.delay = delay
        self.syringe = syringe

    def run(self):
        # Need to 'enable' continuously for keep-alive
        while not self.stopped.wait(self.delay):
            self.enable()
        self.disable()

    def enable(self):
        s = self.syringe
        s.execCommand(Command.remotectrl, [b'ENABLED', s.securitycode])
        s.execCommand(Command.remotecfg, [b'ENABLED', s.securitycode])

    def disable(self):
        s = self.syringe
        s.execCommand(Command.remotectrl, [b'DISABLED'])
        s.execCommand(Command.remotecfg, [b'DISABLED'])

class AlarisSyringe(Syringe):
    def __init__(self, comm):
        super().__init__()
        self.comm = comm
        self.__seccode = None
        self.__kastopper = threading.Event()
        self.launchKeepAlive()

    def __del__(self):
        self.stopKeepAlive()

    def execRawCommand(self, msg, retry=True):
        def qTimeout():
            self.comm.recvq.put(Reply(error = True, value = Error.ETIMEOUT))
            self.comm.allowNewCmd()

        cmd = genFrame(msg)
        self.comm.cmdq.put(cmd)

        # Time out after .5 seconds in case we get no reply.
        t = threading.Timer(.5, qTimeout)
        t.start()
        self.comm.cmdq.join()
        t.cancel()

        reply = self.comm.recvq.get()
        if reply.error and retry and reply.value is Error.ETIMEOUT:
            # Temporary error. Try once more
            printerr("Error: {}. Retrying command.", reply.value)
            return self.execRawCommand(msg, retry=False)
        else:
            return reply

    def execCommand(self, command, fields=[]):
        cmdfields = [command.value] + fields
        commandraw = b'^'.join(cmdfields)
        return self.execRawCommand(commandraw)

    def launchKeepAlive(self):
        looper = Looper(self, delay=1, stopevent=self.__kastopper)
        looper.start()

    def stopKeepAlive(self):
        self.__kastopper.set()

    @property
    def securitycode(self):
        if self.__seccode is None:
            reply = self.execCommand(Command.getserialno)
            if reply.error:
                raise CommandError(reply.value)
            self.__seccode = genCheckSum(reply.value)
        return self.__seccode

    def readRate(self):
        reply = self.execCommand(Command.rate)
        if reply.error:
            raise CommandError(reply.value)
        return reply.value

    def readVolume(self):
        reply = self.execCommand(Command.queryvolume)
        if reply.error:
            raise CommandError(reply.value)
        return reply.value

    def setRate(self, newrate):
        brate = str(newrate).encode('ASCII')
        fields = [brate, b'ml/h']
        reply = self.execCommand(Command.rate, fields)
        if reply.error:
            raise CommandError(reply.value)
        return reply.value

class AlarisComm(serial.Serial):
    def __init__(self, port, baudrate = 38400):
        # These settings come from Alaris documentation
        super().__init__(port     = port,
                         baudrate = baudrate,
                         bytesize = serial.EIGHTBITS,
                         parity   = serial.PARITY_NONE,
                         stopbits = serial.STOPBITS_ONE)
        if DEBUG:
            self.logfile = open('alaris_raw.log', 'wb')

        self.recvq = queue.LifoQueue()
        self.cmdq  = queue.Queue(maxsize = 10)

        self.__rxthread = RecvThread(comm = self)
        self.__txthread = SendThread(comm = self)

        self.__rxthread.start()
        self.__txthread.start()

    def allowNewCmd(self):
        try:
            self.cmdq.task_done()
        except ValueError as e:
            printerr("State machine got confused: {}", e)

    if DEBUG:
        # Write all data exchange to file
        def read(self, size=1):
            data = super().read(size)
            self.logfile.write(data)
            return data

        def write(self, data):
            self.logfile.write(data)
            return super().write(data)

class RecvThread(threading.Thread):
    def __init__(self, comm):
        super().__init__(daemon=True)
        self.comm   = comm
        self.__buffer = b''

    def processRxBuffer(self):
        fields, _ = parseReply(self.__buffer)
        self.__buffer = b''
        reply = Reply(b' '.join(fields))
        self.comm.recvq.put(reply)
        self.comm.allowNewCmd()

    def run(self):
        insideCommand = False
        while True:
            c = self.comm.read(1)
            if c == b'\x1B':
                # Premature termination, b'\x1B' = ESC
                self.__buffer = b''
                insideCommand = False
            elif c == b'!':
                # Start of command marker
                insideCommand = True
            elif c == b'\r':
                # End of command marker
                insideCommand = False
                self.processRxBuffer()
            elif insideCommand:
                self.__buffer += c
            elif c == b'':
                pass
            else:
                printerr("Unexpected char received: {}", ord(c))

class SendThread(threading.Thread):
    def __init__(self, comm):
        super().__init__(daemon=True)
        self.comm = comm

    def run(self):
        while True:
            msg = self.comm.cmdq.get()
            self.comm.write(msg)

class Reply(object):
    __slots__ = ('value', 'error')
    def __init__(self, value = '', error = False):
        self.value = value
        self.error = error

    def __repr__(self):
        return "Alaris Reply: Value={}, Error={}".format(self.value, self.error)

class Command(Enum):
    getserialno = b'INST_SERIALNO'
    remotecfg   = b'REMOTE_CFG'
    remotectrl  = b'REMOTE_CTRL'
    queryvolume = b'INF_VI'
    rate        = b'INF_RATE'
    infstart    = b'INF_START'
    infstop     = b'INF_STOP'

# Errors
class Error(Enum):
    ETIMEOUT = auto()
    EUNDEF  = auto()



import threading
import queue
import time

from enum import Enum, unique
from datetime import datetime

import serial

from infupy.backends.common import Syringe, CommandError, printerr

DEBUG = False

def genCheckSum(msg):
    asciisum = sum(msg)
    _, low = divmod(asciisum, 0x100)
    checksum = 0xFF - low
    checkbytes = b'%02X' % checksum
    return checkbytes

def genFrame(msg):
    return STX + msg + genCheckSum(msg) + ETX

def parseReply(rxbytes):
    # The checksum is in the last two bytes
    chk   = rxbytes[-2:]
    rxmsg = rxbytes[:-2]

    # Partition the string
    splt = rxmsg.split(b';', 1)
    meta = splt[0]
    if len(splt) > 1:
        msg = splt[1]
    else:
        msg = None

    # Read meta data
    if len(meta) > 1:
        origin = meta[0:1]
        status = meta[1:2]
    else:
        origin = None
        status = meta[0:1]

    try:
        restat = ReplyStatus(status)
    except ValueError:
        restat = ReplyStatus.incorrect

    return (restat, origin, msg, chk == genCheckSum(rxmsg))

def parseVars(msg):
    ret = dict()
    if msg is None:
        return ret
    for repvar in msg.split(b';'):
        idbytes = repvar[0:1]
        value = repvar[1:]
        try:
            ident = VarId(idbytes)
        except ValueError:
            continue
        ret[ident] = value
    return ret

def extractRate(msg):
    vals = parseVars(msg)
    if not VarId.rate in vals.keys():
        raise ValueError
    rate = vals[VarId.rate]
    n = int(rate, 16)
    return round(n * 1e-1, 1)

def extractVolume(msg):
    vals = parseVars(msg)
    if not VarId.volume in vals.keys():
        raise ValueError
    vol = vals[VarId.volume]
    n = int(vol, 16)
    return round(n * 1e-3, 3)

class FreseniusModule(Syringe):
    def __init__(self, comm, index=None):
        super().__init__()
        self.comm = comm
        if index is None:
            # Standalone syringe
            index = b''
        if isinstance(index, bytes):
            self._index = index
        else:
            self._index = str(index).encode('ASCII')
        self.connect()

    def execRawCommand(self, msg, retry=True):
        def qTimeout():
            self.comm.recvq.put(Reply(error=True, value=Error.ETIMEOUT))
            self.comm.allowNewCmd()

        cmd = genFrame(self._index + msg)
        self.comm.cmdq.put(cmd)

        # Time out after 1 second in case of communication failure.
        t = threading.Timer(1, qTimeout)
        t.start()
        self.comm.cmdq.join()
        t.cancel()

        reply = self.comm.recvq.get()
        if reply.error and reply.value is Error.ECOMMODULE:
            printerr("Error: {}. Lost connection. Trying to reconnect.", reply.value)
            self.connect()
            return self.execRawCommand(msg, retry=False)
        elif reply.error and retry and reply.value in [Error.ERNR, Error.ETIMEOUT]:
            # Temporary error. Try once more
            printerr("Error: {}. Retrying command.", reply.value)
            return self.execRawCommand(msg, retry=False)
        else:
            return reply

    def execCommand(self, command, flags=[], args=[]):
        if len(flags) > 0:
            flagvals = [f.value for f in flags]
            flagbytes = b''.join(flagvals)
            commandraw = command.value + b';' + flagbytes
        elif len(args) > 0:
            argbytes = b';'.join(args)
            commandraw = command.value + b';' + argbytes
        else:
            commandraw = command.value
        return self.execRawCommand(commandraw)

    def connect(self):
        reply = self.execCommand(Command.connect)
        if reply.error:
            raise CommandError(reply.value)
        return reply.value

    def disconnect(self):
        self.execCommand(Command.disconnect)

    def readDeviceType(self):
        reply = self.execCommand(Command.readfixed, flags=[FixedVarId.devicetype])
        if reply.error:
            raise CommandError(reply.value)
        return reply.value

class FreseniusBase(FreseniusModule):
    def __init__(self, comm, wait=True):
        super().__init__(comm, 0)
        self.syringes = {}
        if wait:
            time.sleep(1)

    def connectSyringe(self, index):
        s = FreseniusSyringe(self.comm, index)
        self.syringes[index] = s
        return s

    def listModules(self):
        modules = []
        reply = self.execCommand(Command.readvar, flags=[VarId.modules])
        if reply.error:
            raise CommandError(reply.value)
        results = parseVars(reply.value)
        binmods = int(results[VarId.modules], 16)
        for i in range(5):
            if (1 << i) & binmods:
                modules.append(i + 1)
        return modules

    def readVolume(self):
        raise NotImplementedError

    def readRate(self):
        raise NotImplementedError

class FreseniusSyringe(FreseniusModule):
    def readRate(self):
        reply = self.execCommand(Command.readvar, flags=[VarId.rate])
        if reply.error:
            raise CommandError(reply.value)
        return extractRate(reply.value)

    def readVolume(self):
        reply = self.execCommand(Command.readvar, flags=[VarId.volume])
        if reply.error:
            raise CommandError(reply.value)
        return extractVolume(reply.value)

    def readDrug(self):
        reply = self.execCommand(Command.readdrug)
        if reply.error:
            raise CommandError(reply.value)
        return reply.value

    def resetVolume(self):
        reply = self.execCommand(Command.resetvolume)
        if reply.error:
            raise CommandError(reply.value)

    def remoteControl(self, doctrl):
        #1 : manual, 2 : PC control
        if doctrl:
            newmode = b'2'
        else:
            newmode = b'1'
        reply = self.execCommand(Command.mode, args=[newmode])
        if reply.error:
            raise CommandError(reply.value)

    def stopInfusion(self):
        reply = self.execCommand(Command.setpause)
        if reply.error:
            raise CommandError(reply.value)

    # Spontaneous variable handling
    def registerEvent(self, event):
        super().registerEvent(event)
        reply = self.execCommand(Command.enspont, flags=self._events)
        if reply.error:
            raise CommandError(reply.value)

    def unregisterEvent(self, event):
        super().unregisterEvent(event)
        reply = self.execCommand(Command.disspont)
        if reply.error:
            raise CommandError(reply.value)
        reply = self.execCommand(Command.enspont, flags=self._events)
        if reply.error:
            raise CommandError(reply.value)

    def clearEvents(self):
        super().clearEvents()
        reply = self.execCommand(Command.disspont)
        if reply.error:
            raise CommandError(reply.value)

    @property
    def index(self):
        return int(self._index)

class FreseniusComm(serial.Serial):
    def __init__(self, port, baudrate=19200):
        # These settings come from Fresenius documentation
        super().__init__(port     = port,
                         baudrate = baudrate,
                         bytesize = serial.SEVENBITS,
                         parity   = serial.PARITY_EVEN,
                         stopbits = serial.STOPBITS_ONE)
        if DEBUG:
            self.logfile = open('fresenius_raw.log', 'wb')

        self.recvq  = queue.LifoQueue()
        self.cmdq   = queue.Queue(maxsize=10)
        self.eventq = queue.Queue(maxsize=1e4)

        # Write lock to make sure only one source writes at a time
        self.__rxthread = RecvThread(self)
        self.__txthread = SendThread(self)

        self.__rxthread.start()
        self.__txthread.start()

    if DEBUG:
        # Write all data exchange to file
        def read(self, size=1):
            data = super().read(size)
            self.logfile.write(data)
            return data

        def write(self, data):
            self.logfile.write(data)
            return super().write(data)

    def allowNewCmd(self):
        try:
            self.cmdq.task_done()
        except ValueError as e:
            printerr("State machine got confused: {}", e)


class RecvThread(threading.Thread):
    def __init__(self, comm):
        super().__init__(daemon=True)
        self.comm = comm
        self.__buffer = b''

    def acknowledgeEvent(self, origin, status):
        self.comm.cmdq.put(genFrame(origin + status.value))
        self.comm.allowNewCmd()

    def enqueueReply(self, reply):
        self.comm.recvq.put(reply)
        self.comm.allowNewCmd()

    def processRxBuffer(self):
        status, origin, msg, chk = parseReply(self.__buffer)
        self.__buffer = b''
        if chk:
            # Send ACK
            self.comm.cmdq.put(ACK)
            self.comm.allowNewCmd()
        else:
            # Send NAK
            printerr("Checksum error: {}", msg)
            self.comm.cmdq.put(NAK + Error.ECHKSUM.value)
            self.comm.allowNewCmd()
            return

        if status is ReplyStatus.incorrect:
            # Error condition
            try:
                error = Error(msg)
            except ValueError:
                error = Error.EUNDEF
            self.enqueueReply(Reply(origin, error, error=True))
            printerr("Command error: {}", error)

        elif status is ReplyStatus.correct:
            # This is a reply to one of our commands
            self.enqueueReply(Reply(origin, msg))

        elif status is ReplyStatus.spont or status is ReplyStatus.spontadj:
            # Spontaneously generated event. We need to acknowledge.
            self.acknowledgeEvent(origin, status)
            if origin is None or not origin.isdigit():
                return
            iorigin = int(origin)
            self.comm.eventq.put((datetime.now(), iorigin, msg))

        else:
            pass

    def run(self):
        insideNAKerr = False
        insideCommand = False
        while True:
            c = self.comm.read(1)
            if c == ENQ:
                # Send keep-alive
                self.comm.cmdq.put(DC4)
                self.comm.allowNewCmd()
            elif insideNAKerr:
                try:
                    error = Error(c)
                except:
                    error = Error.EUNDEF
                self.enqueueReply(Reply(error=True, value=error))
                printerr("Protocol error: {}", error)
                insideNAKerr = False
            elif c == ACK:
                pass
            elif c == STX:
                # Start of command marker
                insideCommand = True
            elif c == ETX:
                # End of command marker
                insideCommand = False
                self.processRxBuffer()
            elif c == NAK:
                insideNAKerr = True
            elif insideCommand:
                self.__buffer += c
            elif c == b'':
                pass
            else:
                printerr("Unexpected char received: {}", ord(c))


class SendThread(threading.Thread):
    def __init__(self, comm):
        super().__init__(daemon=True)
        self.comm = comm

    def run(self):
        while True:
            msg = self.comm.cmdq.get()
            self.comm.write(msg)


# Frame markers
STX = b'\x02'
ETX = b'\x03'

# Delivery control
ACK = b'\x06'
NAK = b'\x15'

# Keep-alive
ENQ = b'\x05'
DC4 = b'\x14'

# Allowed command characters
CHROK = [chr(c) for c in range(0x20 , 0x7E)]

class Reply(object):
    __slots__ = ('origin', 'value', 'error')
    def __init__(self, origin=None, value='', error=False):
        if origin is None:
            self.origin = 0
        else:
            self.origin = int(origin)
        self.value = value
        self.error = error

    def __repr__(self):
        return "Fresenius Reply: Origin={}, Value={}, Error={}".format(self.origin, self.value, self.error)

class Command(Enum):
    connect      = b'DC'
    disconnect   = b'FC'
    mode         = b'MO'
    reset        = b'RZ'
    off          = b'OF'
    silence      = b'SI'
    setdrug      = b'EP'
    readdrug     = b'LP'
    showdrug     = b'AP'
    setid        = b'EN'
    readid       = b'LN'
    enspont      = b'DE'
    disspont     = b'AE'
    readvar      = b'LE'
    enspontadj   = b'DM'
    disspontadj  = b'AM'
    readadj      = b'LM'
    readfixed    = b'LF'
    setrate      = b'PR'
    setpause     = b'PO'
    setbolus     = b'PB'
    setempty     = b'PF'
    setlimvolume = b'PV'
    resetvolume  = b'RV'
    pressurelim  = b'PP'
    dynpressure  = b'PS'

class VarId(Enum):
    alarm   = b'a'
    error   = b'e'
    mode    = b'm'
    rate    = b'd'
    volume  = b'r'
    bolrate = b'k'
    bolvol  = b's'
    nummods = b'i'
    modules = b'b'

class FixedVarId(Enum):
    devicetype = b'b'

class ReplyStatus(Enum):
    correct   = b'C'
    incorrect = b'I'
    spont     = b'E'
    spontadj  = b'M'

# Errors
@unique
class Error(Enum):
    EUNDEF   = b'?'
    # Link layer errors
    ECHAR    = b'\x31'
    ECHKSUM  = b'\x32'
    EADDR    = b'\x34'
    ETIMEOUT = b'\x35'
    ERNR     = b'\x36'
    EFRAME   = b'\x37'
    ECTRL    = b'\x38'
    # Application layer errors
    EUNKNOWN    = b'01'
    ECMDMODE    = b'02'
    ECMDSTAT    = b'03'
    ESYNTAX     = b'04'
    EMODEAUTH   = b'05'
    EMODEAGAIN  = b'06'
    EMODEMODE   = b'07'
    ELIMIT      = b'08'
    EMODESTAT   = b'09'
    EIDENTU     = b'0A'
    EIDENTI     = b'0B'
    EMSGLONG    = b'0C'
    ECOMBASE    = b'0D'
    ECOMMODULEI = b'0E'
    EALARM      = b'12'
    ERATE       = b'14'
    EVOLUME     = b'15'
    EEMPTYMODE  = b'16'
    EEVENT      = b'1A'
    ECOMMODULE  = b'1E'
    ENMAN       = b'1F'
    EPORTAUTH   = b'20'
    ENMODEAUTH  = b'22'
    ECONMODEI   = b'24'
    EDRUG       = b'25'

    def __repr__(self):
        return ERRdescr[self]

ERRdescr = {
    Error.EUNDEF   : "Unknown Error",
    # Link layer errors
    Error.ECHAR    : "Character Reception Problem",
    Error.ECHKSUM  : "Incorrect Check-sum",
    Error.EADDR    : "Incorrect Address",
    Error.ETIMEOUT : "End of [ACK] Character time-out",
    Error.ERNR     : "Receiver not Ready",
    Error.EFRAME   : "Incorrect Frame Length",
    Error.ECTRL    : "Presence of Control Code",
    # Application layer errors
    Error.EUNKNOWN    : "Unknown Command",
    Error.ECMDMODE    : "Command disabled in the current Mode",
    Error.ECMDSTAT    : "Command disabled in this status",
    Error.ESYNTAX     : "Syntax Error",
    Error.EMODEAUTH   : "Operating Mode not Authorized",
    Error.EMODEAGAIN  : "Operating Mode already active",
    Error.EMODEMODE   : "New operating mode disabled in this mode",
    Error.ELIMIT      : "Parameter out off limit",
    Error.EMODESTAT   : "New operating mode disabled in this status",
    Error.EIDENTU     : "Identifier not used",
    Error.EIDENTI     : "Identifier incorrect",                          # a-z
    Error.EMSGLONG    : "Message too long",                              # <= 80
    Error.ECOMBASE    : "Communication session with the base not open",
    Error.ECOMMODULEI : "Communication with module impossible",
    Error.EALARM      : "Presence of an Alarm",
    Error.ERATE       : "Attempt to launch infusion before flow rate selection",
    Error.EVOLUME     : "Insufficient Volume to launch a bolus",
    Error.EEMPTYMODE  : "Impossible to launch the empty Syringe mode",
    Error.EEVENT      : "Recorded event number incorrect",               # 1-64
    Error.ECOMMODULE  : "The Communication with the module is not open",
    Error.ENMAN       : "One of the modules is not in the manual mode",
    Error.EPORTAUTH   : "Command not authorized with this Port",
    Error.ENMODEAUTH  : "New mode unauthorized",
    Error.ECONMODEI   : "Connection Mode incorrect",
    Error.EDRUG       : "Drug number incorrect"
}



import sys
from abc import ABCMeta, abstractmethod

def printerr(msg, e=''):
    msg = "Backend: " + str(msg)
    print(msg.format(e), file=sys.stderr)

class CommandError(Exception):
    def __str__(self):
        return "Command error: {}".format(self.args)

class Syringe(metaclass=ABCMeta):
    _events = set()

    @abstractmethod
    def execCommand(self, msg):
        raise NotImplementedError

    # Read Perfusion related values
    @abstractmethod
    def readRate(self):
        raise NotImplementedError

    @abstractmethod
    def readVolume(self):
        raise NotImplementedError

    # Infusion control
    def setRate(self, rate):
        raise NotImplementedError

    # Events
    def registerEvent(self, event):
        self._events |= set([event])

    def unregisterEvent(self, event):
        self._events -= set([event])

    def clearEvents(self):
        self._events = set()



import sys, os.path, time, csv, io, queue

#from PyQt4 import QtCore, QtWidgets
from qtpy import QtCore, QtWidgets, QtWidgets

import infupy.backends.fresenius as fresenius
from infupy.gui.syringorecueil_ui import Ui_wndMain

DEBUG = True

class Worker(QtCore.QObject):
    sigConnected      = QtCore.Signal()
    sigDisconnected   = QtCore.Signal()
    sigUpdateSyringes = QtCore.Signal(list)
    sigError          = QtCore.Signal(str)

    def __init__(self):
        super(Worker, self).__init__()
        self.oldconnstate = False
        self.destfolder = os.path.expanduser("~")
        self.port = ""
        self.conn = None
        self.base = None
        self.logger = None
        self.syringes = dict()
        self.csvfd = io.IOBase()
        self.csv = None
        self.shouldrun = False

        self.conntimer = QtCore.QTimer()
        self.conntimer.timeout.connect(self.connectionLoop)

        self.logtimer = QtCore.QTimer()
        self.logtimer.timeout.connect(self.logLoop)

        self.conntimer.start(5000) # 5 seconds

    @QtCore.Slot()
    def start(self):
        self.shouldrun = True

    @QtCore.Slot()
    def stop(self):
        self.shouldrun = False

    @QtCore.Slot(str)
    def setport(self, port):
        self.port = port

    @QtCore.Slot(str)
    def setfolder(self, folder):
        self.destfolder = folder

    def connectionLoop(self):
        if not self.shouldrun:
            self.onDisconnected()
            return

        if not self.checkSerial():
            if not self.connectSerial():
                return
        if not self.checkBase():
            if self.connectBase():
                self.onConnected()
            else:
                self.onDisconnected()
                return
        self.checkSyringes()
        self.attachNewSyringes()

    def logLoop(self):
        try: # Ensure file is open and writable.
            if not self.csvfd.writable():
                raise IOError("Not writable")
        except (IOError, ValueError) as e:
            if self.shouldrun:
                self.reportUI("File: {}".format(e))
            return

        while True: # Dump the whole queue to csv
            try:
                dt, origin, msg = self.conn.eventq.get_nowait()
            except queue.Empty:
                break

            try:
                volume = fresenius.extractVolume(msg)
            except ValueError:
                self.reportUI("Failed to decode volume value")
                continue

            if DEBUG: print("{}:{}:{}".format(dt, origin, volume), file=sys.stderr)

            timestamp = int(dt.timestamp() * 1e9)
            self.csv.writerow({'datetime' : timestamp,
                               'syringe'  : origin,
                               'volume'   : volume})

    def onConnected(self):
        if self.oldconnstate == True:
            if DEBUG: print("Already connected", file=sys.stderr)
            return # already connected

        self.oldconnstate = True
        self.sigConnected.emit()
        filename = time.strftime('%Y%m%d-%H%M.csv')
        filepath = os.path.join(self.destfolder, filename)
        if not self.csvfd.writable():
            self.csvfd.close()
            self.csvfd = open(filepath, 'w', newline='')
            self.reportUI("Opened file: {}".format(filepath))
            self.csv = csv.DictWriter(self.csvfd, fieldnames = ['datetime', 'syringe', 'volume'])
            self.csv.writeheader()
        self.logtimer.start(1000) # 1 second

    def onDisconnected(self):
        if self.oldconnstate == False:
            if DEBUG: print("Already disconnected", file=sys.stderr)
            return # already disconnected

        self.oldconnstate = False
        self.sigDisconnected.emit()
        # Clean up
        self.syringes = dict()
        self.base = None
        self.sigUpdateSyringes.emit([])
        # Stop csv logging
        self.logtimer.stop()
        # Call once more to empty the queue
        self.logLoop()
        self.csvfd.close()
        if not self.shouldrun and self.conn is not None:
            self.conn.close()
            self.conn = None

    def checkSyringes(self):
        for i, s in self.syringes.copy().items():
            try:
                dtype = s.readDeviceType()
                if DEBUG: print("Device: {}".format(dtype), file=sys.stderr)
            except Exception as e:
                self.reportUI("Syringe {} lost: {}".format(i, e))
                del self.syringes[i]
            else:
                # Register volume event in case the syringe got reset.
                # If the event was already registered, this is a no-op.
                s.registerEvent(fresenius.VarId.volume)

    def attachNewSyringes(self):
        try:
            modids = self.base.listModules()
            self.sigUpdateSyringes.emit(modids)
            for modid in modids:
                if not modid in self.syringes.keys():
                    s = fresenius.FreseniusSyringe(self.conn, modid)
                    s.registerEvent(fresenius.VarId.volume)
                    self.syringes[modid] = s
        except IOError as e:
            self.reportUI("Attach syringe error: {}".format(e))

    def checkSerial(self):
        if self.conn is None:
            return False
        try:
            self.conn.name
        except Exception as e:
            self.reportUI("Serial port exception: {}".format(e))
            return False
        else:
            return True

    def connectSerial(self):
        try:
            self.conn = fresenius.FreseniusComm(self.port)
        except Exception as e:
            self.reportUI("Failed to open serial port: {}".format(e))
            return False
        else:
            return True

    def checkBase(self):
        if self.base is None:
            return False
        try:
            dtype = self.base.readDeviceType()
            if DEBUG: print("Device: {}".format(dtype), file=sys.stderr)
        except Exception as e:
            self.reportUI("Base error: {}".format(e))
            return False
        else:
            return True

    def connectBase(self):
        try:
            self.base = fresenius.FreseniusBase(self.conn)
        except Exception as e:
            self.reportUI("Failed to connect to base: {}".format(e))
            return False
        else:
            return True

    def reportUI(self, err):
        if DEBUG: print(err, file=sys.stderr)
        self.sigError.emit(str(err))

    @QtCore.Slot()
    def cleanup(self):
        self.logtimer.stop()
        self.conntimer.stop()
        for _, s in self.syringes.items():
            s.disconnect()
        if self.base is not None:
            self.base.disconnect()
        if self.conn is not None:
            self.conn.close()
        if self.csvfd is not None:
            self.csvfd.close()


class MainUi(QtWidgets.QMainWindow, Ui_wndMain):
    sigCleanup = QtCore.Signal()

    def __init__(self, parent = None):
        super(MainUi, self).__init__(parent = parent)
        self.setupUi(self)

        # Add Connection label to statusbar
        self.connStatusLabel = QtWidgets.QLabel()
        self.connStatusLabel.setMargin(2)
        self.statusBar.addPermanentWidget(self.connStatusLabel)

        # Init worker
        self.__workerthread = QtCore.QThread()
        self.__worker = Worker()

        # Worker callbacks and signals
        self.btnStart.clicked.connect(self.__worker.start)
        self.btnStop.clicked.connect(self.__worker.stop)
        self.sigCleanup.connect(self.__worker.cleanup)
        self.txtFolder.textChanged.connect(self.__worker.setfolder)
        self.comboCom.editTextChanged.connect(self.__worker.setport)
        self.__worker.setport(self.comboCom.currentText())

        self.__worker.sigConnected.connect(self.connected)
        self.__worker.sigDisconnected.connect(self.disconnected)
        self.__worker.sigUpdateSyringes.connect(self.updateSyringeList)
        self.__worker.sigError.connect(self.showStatusError)

        self.__worker.moveToThread(self.__workerthread)
        self.__workerthread.start()

        # Continue UI initialization
        self.btnBrowse.clicked.connect(self.browsefolder)

    def browsefolder(self):
        folderpreset = self.txtFolder.text()
        if not os.path.isdir(folderpreset):
            folderpreset = os.path.expanduser("~")
        destfolder = QtWidgets.QFileDialog.getExistingDirectory(self, "Choose destination Folder", folderpreset)
        self.txtFolder.setText(destfolder)

    def showStatusError(self, errstr):
        # Show for 3 seconds
        self.statusBar.showMessage(errstr, 3000)

    def connected(self):
        self.connStatusLabel.setStyleSheet("QLabel{background : green;}")
        self.connStatusLabel.setText("Connected")

    def disconnected(self):
        self.lstSyringes.clear()
        self.connStatusLabel.setStyleSheet("QLabel{background : red;}")
        self.connStatusLabel.setText("Disconnected")

    def updateSyringeList(self, slist):
        self.lstSyringes.clear()
        for modid in slist:
            liststr = "Seringue {}".format(modid)
            self.lstSyringes.addItem(liststr)

    def closeEvent(self, event):
        # Wrap it all up
        if DEBUG: print("Cleaning up before exiting.", file=sys.stderr)
        self.sigCleanup.emit()
        self.__workerthread.quit()
        event.accept()


if __name__ == '__main__':
    qApp = QtWidgets.QApplication(sys.argv)

    wMain = MainUi()
    wMain.show()

    sys.exit(qApp.exec_())



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'syringorecueil.ui'
#
# Created by: PyQt4 UI code generator 4.11.4
#
# WARNING! All changes made in this file will be lost!

#from PyQt4 import QtCore, QtWidgets
from qtpy import QtCore, QtWidgets, QtWidgets

try:
    _fromUtf8 = QtCore.QString.fromUtf8
except AttributeError:
    def _fromUtf8(s):
        return s

try:
    _encoding = QtWidgets.QApplication.UnicodeUTF8
    def _translate(context, text, disambig):
        return QtWidgets.QApplication.translate(context, text, disambig, _encoding)
except AttributeError:
    def _translate(context, text, disambig):
        return QtWidgets.QApplication.translate(context, text, disambig)

class Ui_wndMain(object):
    def setupUi(self, wndMain):
        wndMain.setObjectName(_fromUtf8("wndMain"))
        wndMain.resize(466, 468)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(wndMain.sizePolicy().hasHeightForWidth())
        wndMain.setSizePolicy(sizePolicy)
        self.centralwidget = QtWidgets.QWidget(wndMain)
        self.centralwidget.setObjectName(_fromUtf8("centralwidget"))
        self.verticalLayout = QtWidgets.QVBoxLayout(self.centralwidget)
        self.verticalLayout.setObjectName(_fromUtf8("verticalLayout"))
        self.groupBox_3 = QtWidgets.QGroupBox(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.groupBox_3.sizePolicy().hasHeightForWidth())
        self.groupBox_3.setSizePolicy(sizePolicy)
        self.groupBox_3.setObjectName(_fromUtf8("groupBox_3"))
        self.verticalLayout_4 = QtWidgets.QVBoxLayout(self.groupBox_3)
        self.verticalLayout_4.setObjectName(_fromUtf8("verticalLayout_4"))
        self.comboCom = QtWidgets.QComboBox(self.groupBox_3)
        self.comboCom.setEditable(True)
        self.comboCom.setObjectName(_fromUtf8("comboCom"))
        self.comboCom.addItem(_fromUtf8(""))
        self.comboCom.addItem(_fromUtf8(""))
        self.comboCom.addItem(_fromUtf8(""))
        self.comboCom.addItem(_fromUtf8(""))
        self.comboCom.addItem(_fromUtf8(""))
        self.comboCom.addItem(_fromUtf8(""))
        self.comboCom.addItem(_fromUtf8(""))
        self.comboCom.addItem(_fromUtf8(""))
        self.comboCom.addItem(_fromUtf8(""))
        self.verticalLayout_4.addWidget(self.comboCom)
        self.verticalLayout.addWidget(self.groupBox_3)
        self.groupBox_2 = QtWidgets.QGroupBox(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.groupBox_2.sizePolicy().hasHeightForWidth())
        self.groupBox_2.setSizePolicy(sizePolicy)
        self.groupBox_2.setObjectName(_fromUtf8("groupBox_2"))
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.groupBox_2)
        self.verticalLayout_3.setObjectName(_fromUtf8("verticalLayout_3"))
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName(_fromUtf8("horizontalLayout"))
        self.txtFolder = QtWidgets.QLineEdit(self.groupBox_2)
        self.txtFolder.setObjectName(_fromUtf8("txtFolder"))
        self.horizontalLayout.addWidget(self.txtFolder)
        self.btnBrowse = QtWidgets.QPushButton(self.groupBox_2)
        self.btnBrowse.setObjectName(_fromUtf8("btnBrowse"))
        self.horizontalLayout.addWidget(self.btnBrowse)
        self.verticalLayout_3.addLayout(self.horizontalLayout)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName(_fromUtf8("horizontalLayout_2"))
        self.btnStart = QtWidgets.QPushButton(self.groupBox_2)
        self.btnStart.setObjectName(_fromUtf8("btnStart"))
        self.horizontalLayout_2.addWidget(self.btnStart)
        self.btnStop = QtWidgets.QPushButton(self.groupBox_2)
        self.btnStop.setObjectName(_fromUtf8("btnStop"))
        self.horizontalLayout_2.addWidget(self.btnStop)
        self.verticalLayout_3.addLayout(self.horizontalLayout_2)
        self.verticalLayout.addWidget(self.groupBox_2)
        self.groupBox = QtWidgets.QGroupBox(self.centralwidget)
        self.groupBox.setObjectName(_fromUtf8("groupBox"))
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.groupBox)
        self.verticalLayout_2.setObjectName(_fromUtf8("verticalLayout_2"))
        self.lstSyringes = QtWidgets.QListWidget(self.groupBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.lstSyringes.sizePolicy().hasHeightForWidth())
        self.lstSyringes.setSizePolicy(sizePolicy)
        self.lstSyringes.setObjectName(_fromUtf8("lstSyringes"))
        self.verticalLayout_2.addWidget(self.lstSyringes)
        self.verticalLayout.addWidget(self.groupBox)
        wndMain.setCentralWidget(self.centralwidget)
        self.statusBar = QtWidgets.QStatusBar(wndMain)
        self.statusBar.setObjectName(_fromUtf8("statusBar"))
        wndMain.setStatusBar(self.statusBar)

        self.retranslateUi(wndMain)
        QtCore.QMetaObject.connectSlotsByName(wndMain)

    def retranslateUi(self, wndMain):
        wndMain.setWindowTitle(_translate("wndMain", "SyRe", None))
        self.groupBox_3.setTitle(_translate("wndMain", "Serial Port", None))
        self.comboCom.setItemText(0, _translate("wndMain", "COM1", None))
        self.comboCom.setItemText(1, _translate("wndMain", "COM2", None))
        self.comboCom.setItemText(2, _translate("wndMain", "COM3", None))
        self.comboCom.setItemText(3, _translate("wndMain", "COM4", None))
        self.comboCom.setItemText(4, _translate("wndMain", "COM5", None))
        self.comboCom.setItemText(5, _translate("wndMain", "COM6", None))
        self.comboCom.setItemText(6, _translate("wndMain", "COM7", None))
        self.comboCom.setItemText(7, _translate("wndMain", "/dev/ttyUSB0", None))
        self.comboCom.setItemText(8, _translate("wndMain", "/dev/ttyUSB1", None))
        self.groupBox_2.setTitle(_translate("wndMain", "Destination Folder", None))
        self.btnBrowse.setText(_translate("wndMain", "Browse", None))
        self.btnStart.setText(_translate("wndMain", "Start", None))
        self.btnStop.setText(_translate("wndMain", "Stop", None))
        self.groupBox.setTitle(_translate("wndMain", "Syringe Pumps", None))




from setuptools import setup

setup(name         = 'infupy',
      version      = '0.3',
      description  = 'Syringe pump infusion',
      url          = 'https://github.com/jaj42/infupy',
      author       = 'Jona Joachim',
      author_email = 'jona@joachim.cc',
      license      = 'ISC',
      packages     = ['infupy', 'infupy.backends', 'infupy.gui'],
      install_requires=[
          'pyserial',
          'crcmod',
          'qtpy'
      ],
      scripts = [
          'scripts/syre.pyw'
      ]
)
